// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:nil -*-
/* 
 * Copyright (C) 2011 Micha Hersch,
 *
 * Permission is granted to copy, distribute, and/or modify this program
 * under the terms of the GNU General Public License, version 2 or any
 * later version published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details
 */
#ifndef __RIGID_TRANSFO_H__
#define __RIGID_TRANSFO_H__

#include "mathlib.h"
#include "Translation.h"
#include "Rotation.h"




/**
 * @brief represents a rigid transformation using the Euler (or Rodrigues)
 * parametrization
 * for rotation. See Hestenes, "New Foundations for Classical Mechanics",
 * pp 280 for the mathematical formalism and Hersch, Billard, Bergmann,
 * "Iterative Estimation of Rigid-Body Transformation - Application to
 * Robust Object Tracking and Iterative Closest Point", Journal of Mathematical
 * Imaging and Vision, 2011. 
 */
class RigidTransfo : public Rotation, public Translation{
   
 protected:
  int adaptive;

 public:
   RigidTransfo(){adaptive=1;};
  
    /**
     * @brief performs a gradient descent on the beta parameter of
     * the rotation and on the translation in order to have a rigid
     * transformation that brings v closer
     * to v_tr. See Hersch, Billard, Bergmann,"Iterative Estimation of 
     * Rigid-Body Transformation - Application to
     * Robust Object Tracking and Iterative Closest Point",
     * Journal of Mathematical Imaging and Vision, 2011.  
     * Only the direction of the gradient of beta is taken
     * into account, not its size.
     * @param v vector a point that can be transformed
     * @param v_tr the observed image of v
     */
   float Update(CVector3_t v, CVector3_t v_tr);


  int SetTransfo(CVector3_t beta, CVector3_t transl);
 

 /**
  * @brief sets the learnig rate sizes
  * @param new_eps rotation step size
  * @param new_eps_tr translation step size
  */
  virtual void SetRate(float new_eps,float new_eps_tr );

  /** 
   * @briefscale the learning rate sizes 
   * @param factor the scaling factor (for both rotation and
   * translation learning ratea)
   */
  virtual void ScaleRate(float factor);

  /**
   * @brief Performs the rigid tranformation
   * @param in the vector to be transformed
   * @param out the resulting (transformed) vector
   */
  void Transform(const CVector3_t in, CVector3_t out);


  void Transform(const CVector3_t in, float angle, CVector3_t out)
  {SetAngle(angle);Transform(in,out);};
  /**
   * @brief Performs the inverse rigid tranformation
   * @param in the vector to be transformed
   * @param out the resulting (transformed) vector
   */

   void InverseTransform(const CVector3_t in,CVector3_t out);
   void ReverseTransform(const CVector3_t in,CVector3_t out);

   void InverseTransform(CVector3_t in, float angle,CVector3_t out)
   {SetAngle(angle);InverseTransform(in,out);};

   void Invert();
   void NoLearning(){adaptive=0;};
   int IsAdaptive()const{return adaptive;}
   void Copy(const RigidTransfo& r){Rotation::Copy(r);Translation::Copy(r);adaptive=r.IsAdaptive();}
   RigidTransfo& operator*(const RigidTransfo& rt1);
   void Identity(){v_clear(beta);alpha=1;v_clear(trans);}
};

/**
 * @brief output streaming operator
 */
ostream& operator<<(ostream& out, const RigidTransfo& rt);

/**
 * @brief input streaming operator
 */
istream& operator>>(istream& in, RigidTransfo& rt);
#endif
