// -*- mode:C++; tab-width:4; c-basic-offset:4; indent-tabs-mode:nil -*-
/* 
 * Copyright (C) 2011 Micha Hersch,
 *
 * Permission is granted to copy, distribute, and/or modify this program
 * under the terms of the GNU General Public License, version 2 or any
 * later version published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General
 * Public License for more details
 */

#include <algorithm>
#include <string.h>
#include "mathlib.h"
#include "icp.h"
#include "RigidTransfo.h"

using namespace std;

RigidTransfoBuffer::RigidTransfoBuffer(int s){
  size=s;
  buffer = new RigidTransfo[size];
  top=bottom=nb=0;
}

int RigidTransfoBuffer::addTransfo(RigidTransfo& rt){
  if(nb==size){
    bottom = (bottom+1)%size;
  }
  else{
    nb++;
  }
  buffer[top]=rt;
  top =  (top+1)%size;
  return nb;
}


float RigidTransfoBuffer::getSlope(){

  CVector3_t rot_num,trans_num,v;
  int cnt;
  float den=0;
  float in=(nb-1)/2.0;
  //in = in/nb;
  v_clear(rot_num);
  v_clear(trans_num);
  if(nb<1) return -1.0;
  for(int i=0;i<nb;i++){
    cnt= (top+i)%size;
    buffer[cnt].GetTranslation(v);
    v_scale(v,i-in,v);
    v_add(v,trans_num,trans_num);
    buffer[cnt].GetRotationParam(v);
    v_scale(v,i-in,v);
    v_add(v,rot_num,rot_num);
    den+=-in*i+i*i;
 }
 den = 1.0f/den;
 v_scale(rot_num,den,rot_num);
 v_scale(trans_num,den,trans_num);
 return sqrtf(v_squ_length(trans_num)+v_squ_length(rot_num));
}



ICP::ICP(){
  ndata =ntdata=0;
  cand.SetRate(0.0005,0.0005); //default learning rate
  show_traj=0;
  data = tdata=cdata=NULL;
}

ICP::~ICP(){
  // if(data) delete[] data;
  // if(tdata) delete[] tdata;
  //  if(cdata) delete[] cdata;
}





float* ICP::generate_data(int n){
 
  float *data = new float[3*n];
  ndata =n;
  float sumx,sumy,sumz;
  sumx=sumy=sumz=0;
  for(int i=0;i<n;i++){
    int ind= 3*i; 
    float x = RND(10);
    float y = RND(10);
//     sumx += data[ind] = x*y/5;
//     sumy += data[ind+1]= sqrtf((x-5)*(x-5)+(y-5)*(y-5));
//     sumz += data[ind+2]= RND(2);   
    sumx += data[ind] = x;
    sumy += data[ind+1]= y;
    sumz += data[ind+2]= 0;   

  }
#ifdef CENTERING
  // data centering
  sumx = sumx/n;
  sumy = sumy/n;
  sumz = sumz/n;

  for(int i=0;i<n;i++){
    int ind= 3*i; 
    data[ind]-=sumx;
    data[ind+1]-=sumy;
    data[ind+2]-=sumz;
  }
#endif
  return data;  
}

void ICP::center_data(float *data,int n,CVector3_t mean){
  float sumx,sumy,sumz;
  sumx=sumy=sumz=0;
  for(int i=0;i<n;i++){ 
    int ind= 3*i; 
    sumx+=data[ind];
    sumy+=data[ind+1];
    sumz+=data[ind+2];
  }
 sumx/=n;
  sumy/=n;
  sumz/=n; 
 for(int i=0;i<n;i++){
    int ind= 3*i; 
    data[ind]-=sumx;
    data[ind+1]-=sumy;
    data[ind+2]-=sumz;
  }
 
  v_set(sumx,sumy,sumz,mean);
}
 

void ICP::translate_data(float *mdata,int n,CVector3_t trans){
  for(int i=0;i<n;i++){
    int ind= 3*i; 
    mdata[ind]+=trans[0];
    mdata[ind+1]+=trans[1];
    mdata[ind+2]+=trans[2];
  }
} 

float *ICP::read_data_from_file(const char *filename, int& n){
  ifstream infile(filename);
  if(!infile.is_open()){return NULL;}
  char line[100];
  float x,y,z;
  int step;
  int nlines = count(istreambuf_iterator<char>(infile),
                      istreambuf_iterator<char>(),'\n');
  if(!n){
    step=1;
    n=nlines;
  }
  else{ 
    step = nlines/n;
  }
  float *mdata = new float[3*n];

  infile.seekg(0, ios::beg);

  int i= 0;
  while(!infile.eof() && i<n){
    if(rand()%step){
      infile.getline(line,100);
    }
    else{
      int ind = 3*i;
      infile>>x>>y>>z;
      mdata[ind]=x;mdata[ind+1]=y;mdata[ind+2]=z;
      i++;
    }
  }
  n=i;
  infile.close();
  return mdata;
}

void ICP::random_transfo(RigidTransfo& rt, float max_angle){
  CVector3_t trans;
   CVector3_t rot;
   float angle = RND(max(0,min(pi,max_angle)));
   v_set(RND(5),RND(5),RND(5),trans);
   v_set(RND(2)-1,RND(2)-1,RND(2)-1,rot);
   v_normalize(rot,rot);
   

   v_scale(rot,0.5,rot);
   rt.SetTransfo(rot,trans);
   rt.SetAngle(angle);
}


float* ICP::transform_data(RigidTransfo& rt){
  CVector3_t v1;
  int n = ndata;
  float *dest = new float[3*n];
  
   for(int i=0;i<n;i++){
     int ind= 3*i;
     rt.Transform(data+ind,dest+ind); //no noise for now
     //with noise
     v_set(normalSample(0,0.01),normalSample(0,0.01),normalSample(0,0.01),v1);    
     v_add(v1,dest+ind,dest+ind);
   }
   return dest;
}

void ICP::print_data(float *data, int n, const char *fname=NULL){
  ostream *out;
  if(!fname){
    out = &cout;
  }
  else{
    out = new ofstream(fname);
  }
  for(int i=0;i<n;i++){
    int ind= 3*i; 
    (*out)<<data[ind]<<" "<<data[ind+1]<<" "<<data[ind+2]<<endl;
  }
  if(fname){
    ((ofstream *)out)->close();
    delete out;
  }
}

void ICP::init(){
  srand(time(NULL));
}


int ICP::find_closest_point(float *point, float *data, int n){
 
  float d2,mmax = FLT_MAX;
  CVector3_t diff;
  int ind = 0;
  for(int i=0;i<n;i++){
    v_sub(point,data+3*i,diff);
    d2 = v_squ_length(diff);
    if(d2<mmax){
      mmax=d2;
      ind = i;
    }
  }
  return 3*ind;
}

int ICP::perform_icp(int nbit){
  CVector3_t p1,p2;
  // CQuat_t q1,q2,q3;
  int ind,cn;
  int n1 = ndata;
  int n2 = ntdata;
  int i;
  for(i=0;i<nbit;i++){

    if(1){      
       ind = 3*(rand()%n1);    
      v_copy(data+ind,p1);
      cand.Transform(p1,p2);
      cn = find_closest_point(p2,tdata,ntdata);
      v_copy(tdata+cn,p2);
      cand.Update(p1,p2);
    }
    else{//back transfo
       ind = 3*(rand()%n2);    
      v_copy(tdata+ind,p2);
      cand.InverseTransform(p2,p1);
      cn = find_closest_point(p1,data,ndata);
      v_copy(data+cn,p1);
      cand.Update(p1,p2);
    }
    if(!(i%100)){
      if(show_traj){
	cout<<cand<<endl;
      }
    if(trace.addTransfo(cand)==trace.getSize()){
      float slope = trace.getSlope();
      if (slope<0.00001){
	return i;
      }
    }
    }
  }
  return i;
}

/**
 * updates the translation term to have a rotation centered in zero
 */
void ICP::decenter_transform(){
  CVector3_t v1,v2;
  cand.Rotation::Transform(data_center,v1);
  v_sub(data_center,v1,v2);
  cand.GetTranslation(v1);
  v_add(v2,v1,v2);
  cand.SetTranslation(v2);
}



void ICP::icp_from_files(char *fname1, char *fname2, int maxit){
  CVector3_t mean,trans;
  CMatrix3_t rot;
  init();
  data = read_data_from_file(fname1,ndata);
  tdata = read_data_from_file(fname2,ntdata);
  center_data(data,ndata,data_center);
  v_scale(data_center,-1,mean);
  translate_data(tdata,ntdata,mean);
  cand.Identity();
  perform_icp(maxit);
  decenter_transform();
  cand.RotToMatrix(rot);
  cand.GetTranslation(trans);
  v_copy(trans,rot+12);
  cout.width(10);
  cout<<cand<<endl;
  //uncomment this if you want to print the homogeneous matrix 
  //   cout<<"transformation matrix"<<endl;  
  //   cout<<rot<<endl;
  delete[]data;
  delete[]tdata;
}


void ICP::generate_data_files(const char *fname1, const char *fname2, int nb){
  RigidTransfo rt;
  CVector3_t trans;
  CMatrix3_t rot;

  init();
  data = generate_data(nb);
  random_transfo(rt,pi/18);
  tdata= transform_data(rt);
  print_data(data,nb,fname1);
  print_data(tdata,nb,fname2);

  rt.RotToMatrix(rot);
  rt.GetTranslation(trans);
  v_copy(trans,rot+12);
  cout.width(10);
  
  cout<< rt<<endl;
  // cout<<rot<<endl; //uncomment this if you want to print the homogeneous matrix 
  delete[]data;
  delete[]tdata;

}

#ifdef GENERATE_DATA
int main(int argc, char *argv[]){
  ICP algo;

  if(argc<4){
    cout<<"usage: icp <file1> <file2> <nbPoints>"<<endl;
    return 0;
  }
  algo.generate_data_files(argv[1],argv[2],atoi(argv[3]));
  return 0;
}

#else
int main(int argc, char *argv[]){
  ICP algo;
  if(argc<4){
    cout<<"usage: icp <file1> <file2> <maxIterationNumber> [--verbose]"<<endl;
    return 0;
  }
  if(argc ==5){
    if(!strcmp(argv[4],"--verbose")){
	algo.set_verbose();
      }
  }
  algo.icp_from_files(argv[1],argv[2],atoi(argv[3]));
}
#endif
